/* src.c - sample rate conversion */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#ifdef MSC_VER
#include <malloc.h>
#else
#define _access access
#endif /* MSC_VER */

#ifndef SIGPTRDEF
#ifdef HUGE_PTR
typedef short huge *SIGPTR;
#else
typedef short *SIGPTR;
#endif /* HUGE_PTR */
#define SIGPTRDEF
#endif  /* SIGPTRDEF */

#define MAXUPSR   	512             /* maximum upsample ratio */
#define BUFFER_SIZE	4096            /* input & output buffer sizes */
#define limit(a,b,c)    (((b)<(a))?(a):((b)>(c))?(c):(b))
#ifndef M_PI
#define M_PI 3.14159265358979323846264338327950288
#endif

/* data variables and arrays */
static double rr;			/* rate ratio = rate_out/rate_in */
static float   scale = 1.0;             /* gain */
static float   ofst;			/* scaled offset */
static float  *hh = NULL;		/* low-pass filter function */
static int upsr = 0;		        /* upsample ratio */
static int fltwid = 16;	                /* filter width */
static int n_up;		        /* upsample ratio */
static int n_fw;		        /* filter width */
static long rate_out = 0;		/* output sampling rate desired */
static long rate_in;			/* read in sampling rate */
static long nsmp_in = 0, nsmp_out = 0;

/* buffers and related indicators */
static int buf_in_over = 0;             /* input buffer overrun indicator */
static long buf_in_head, buf_in_end;	/* input buffer indicators */
static long buf_in_pfw, buf_in_mfw;	/* input buffer pos +/- filter width */
static short buf_len;	                /* # of samples in buffer */
static short *buf_in = NULL;		/* input buffer in int */
static short *buf_out = NULL;		/* output buffer */

static void
rationalize()
{
    int i1, i2, o1, o2;
    double irat, dif, min;

    i1 = i2 = o1 = o2 = 1;
    min = dif = fabs(rr - 1);
    while(i1 < MAXUPSR && i2 < MAXUPSR) {
	irat = (double) i1 / (double) i2;
	dif = fabs(rr - irat);
	if (min > dif) {
	    min = dif;
	    o1 = i1;
	    o2 = i2;
	}
	if (irat < rr) {
	    i1++;
	} else if(irat > rr) {
	    i2++;
	} else {
	    break;
	}
    }
    if (upsr <= 0) {
	n_up = o1;
	upsr = (o1 > o2) ? o1 : o2;
    } else if (rate_out >= rate_in) {
	n_up = upsr;
    } else {
	n_up = (int) (1 + upsr * rr);
    }
    n_fw = (upsr * fltwid) / n_up;
    if (n_fw > BUFFER_SIZE / 2)
        n_fw = BUFFER_SIZE / 2;
}

static int
mem_alloc(void)
{
    int     nn;

    nn = upsr * fltwid + 1;
    if (hh != NULL)
        free(hh);
    if ((hh = (float *) calloc(nn, sizeof(float))) == NULL)
        return(0);
    if (buf_in == NULL)
        if ((buf_in = (short *) malloc(BUFFER_SIZE)) == NULL)
            return(0);
    return (1);
}

static void
initialize()
{
    double  a, p, w, sum;
    int     nn, i, j, k, l;
    float   s, h;

    nn = upsr * fltwid + 1;
    hh[0] = 1;
    for (i = 1; i < nn; i++) {
	a = (double) i * M_PI / upsr;
	p = (double) i * M_PI / (nn - 1);
	w = 0.42 + 0.5 * cos(p) + 0.08 * cos(2 * p);	/* Blackman window  */
	h = (float) (w * sin(a) / a);
	hh[i] = h;			/* LPF impulse response */
    }

    for (k = 0; k <= (n_up / 2); k++) {	/* normalize */
	sum = 0;
	for (i = -n_fw; i < n_fw; i++) {
	    j = (i < 0) ? -(i * n_up + k) : (i * n_up + k);
	    sum += hh[j];
	}
	s = (float) (scale / sum);
	l = (k == 0) ? 1 : ((k == (n_up - k))) ? 0 : n_fw;
	for (i = -n_fw; i < l; i++) {
	    j = (i < 0) ? -(i * n_up + k) : (i * n_up + k);
	    hh[j] *= s;
	}
    }

}

int
chk_buf_in(long i)
{
    buf_in_mfw = i / n_up - n_fw;
    buf_in_pfw = i / n_up + n_fw;
    buf_in_over = (buf_in_mfw < buf_in_head || buf_in_pfw >= buf_in_end);
    return (buf_in_over);
}

static short
interp(long i)
{
    double  sum, v;
    int     ii, jj, k2;
    long    k1;
    SIGPTR  ff;

    k1 = i / n_up;
    ff = buf_in + (k1 - buf_in_head);

    k2 = n_up - (int) (i % n_up);
    sum = ofst;
    for (ii = -n_fw; ii < n_fw; ii++) {
	jj = (ii < 0) ? -(ii * n_up + k2) : (ii * n_up + k2);
	sum += hh[jj] * ff[ii];
    }
    v = floor(sum + 0.5);
    return ((short) limit(-32768, v, 32767));
}

int
src_init(long i_np, long i_rt, long o_rt, int fw, double gn, long *o_np)
{
    int i;
    static long ri = 0, ro = 0;

    if (i_np <= 0 || i_rt <= 0)
        return (0);
    nsmp_in = i_np;
    rate_in = i_rt;
    rate_out = o_rt;
    fltwid = fw;
    scale = (float) gn;
    rr = (double) rate_out / (double) rate_in;
    nsmp_out = (long) (nsmp_in * rr + 0.5);
    if (rate_in != ri || rate_out != ro || hh == NULL) {
        rationalize();
        if (!mem_alloc())		/* allocate memory */
            return (0);
        initialize();		/* initilize buffers and etc. */
        ri = rate_in;
        ro = rate_out;
    }
    /* initialize input buffer */
    buf_len = BUFFER_SIZE / sizeof(short);
    buf_in_head = -buf_len;
    buf_in_end = 0;
    buf_in_over = 0;
    (void) chk_buf_in((long) -n_fw);
    for (i = buf_len - n_fw; i < buf_len; i++)
        buf_in[i] = 0;
    *o_np = nsmp_out;
    return(nsmp_out != nsmp_in);
}

long
src_push(long np, SIGPTR *buf)
{
    long i, m, n;
    
    if (np <= 0) {
        np = 0;
    } else if (np > 0) {
        m = buf_len + buf_in_mfw - buf_in_end;
        if (np > m)
            np = m;
        n = buf_len - np;
        if (np <= 0)
            return (0);
        for (i = 0; i < n; i++)
            buf_in[i] = buf_in[i + np];
        buf_in_head += np;
        buf_in_end += np;
        buf_in_over = 0;
        *buf = buf_in + n;
    }
    return (np);
}

long
src_put(SIGPTR bf, long np)
{
    long i, n;
    SIGPTR buf;
    
    n = src_push(np, &buf);
    for (i = 0; i < n; i++)
        buf[i] = bf[i];
    return (n);
}

static void
src_eos()
{
    long i, n;
    SIGPTR buf;

    n = src_push((long) n_fw, &buf);
    for (i = 0; i < n; i++)
        buf[i] = 0;
}

long
src_get(long ip, long np, int eos, SIGPTR buf)
{
    long i, j;
    double  sr;

    if (np > buf_len)
        np = buf_len;
    sr = n_up / rr;
    for (j = 0; j < np; j++) {
	i = (long) ((j + ip) * sr + 0.5);
        if (chk_buf_in(i)) {
            if (!eos)
                break;
            src_eos();
       	}
	buf[j] = interp(i);
	if (buf_in_over)
	    break;
    }
    return (j);
}

long
src_pull(long ip, long np, int eos, SIGPTR *buf)
{
    if (buf_out == NULL)
        if ((buf_out = (short *) malloc(BUFFER_SIZE)) == NULL)
            return(0);
    *buf = buf_out;
    return (src_get(ip, np, eos, buf_out));
}

void
src_free()
{
    if (hh != NULL) {
        free(hh);
        hh = NULL;
    }
    if (buf_in != NULL) {
        free(buf_in);
        buf_in = NULL;
    }
    if (buf_out != NULL) {
        free(buf_out);
        buf_out = NULL;
    }
}
